/* $Id: hawki_fwhm.h,v 1.4 2015/08/07 13:07:06 jim Exp $
 *
 * This file is part of the HAWKI Pipeline
 * Copyright (C) 2015 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: jim $
 * $Date: 2015/08/07 13:07:06 $
 * $Revision: 1.4 $
 * $Name:  $
 */


/* Includes */

#ifndef HAWKI_FWHM_H
#define HAWKI_FWHM_H

#include "cpl_propertylist.h"

/* Function prototypes */

double hawki_get_dimm_fwhm(cpl_propertylist * plist);
double hawki_get_fwhm_dimm_filter(double fwhm_dimm, double wlength, 
                                  double airmass);
double hawki_get_filter_central_wave(const char * filter);

#endif
